@@ -33,6 +33,7 @@
 #include <inttypes.h>
 #include <time.h>
 
+#include <sched.h>
 #include <sys/stat.h>
 #include <sys/types.h>
 #include <fcntl.h>
@@ -247,7 +248,7 @@ copy_owner (const char *from, const char *to, libcrun_error_t *err)
 
   ret = get_file_owner (from, &uid, &gid);
   if (UNLIKELY (ret < 0))
-    return crun_make_error (err, errno, "cannot get file owner for %s", from);
+    return crun_make_error (err, errno, "cannot get file owner for `%s`", from);
 
   if (uid == 0 && gid == 0)
     return 0;
@@ -271,13 +272,13 @@ read_unified_cgroup_pid (pid_t pid, char **path, libcrun_error_t *err)
 
   from = strstr (content, "0::");
   if (UNLIKELY (from == NULL))
-    return crun_make_error (err, -1, "cannot find cgroup2 for the process %d", pid);
+    return crun_make_error (err, 0, "cannot find cgroup2 for the process `%d`", pid);
 
   from += 3;
 
   to = strchr (from, '\n');
   if (UNLIKELY (to == NULL))
-    return crun_make_error (err, -1, "cannot parse `%s`", cgroup_path);
+    return crun_make_error (err, 0, "cannot parse `%s`", cgroup_path);
   *to = '\0';
 
   *path = xstrdup (from);
@@ -303,7 +304,7 @@ enter_cgroup_v1 (pid_t pid, const char *path, bool create_if_missing, libcrun_er
   if (UNLIKELY (rootless < 0))
     return rootless;
 
-  ret = read_all_file ("/proc/self/cgroup", &content, &content_size, err);
+  ret = read_all_file (PROC_SELF_CGROUP, &content, &content_size, err);
   if (UNLIKELY (ret < 0))
     {
       if (crun_error_get_errno (err) == ENOENT)
@@ -392,7 +393,7 @@ enter_cgroup_v2 (pid_t pid, pid_t init_pid, const char *path, bool create_if_mis
     {
       crun_error_release (err);
 
-      ret = make_cgroup_threaded (path, err);
+      ret = maybe_make_cgroup_threaded (path, err);
       if (UNLIKELY (ret < 0))
         return ret;
 
@@ -458,8 +459,33 @@ int
 enter_cgroup (int cgroup_mode, pid_t pid, pid_t init_pid, const char *path,
               bool create_if_missing, libcrun_error_t *err)
 {
+  int ret;
   if (cgroup_mode == CGROUP_MODE_UNIFIED)
-    return enter_cgroup_v2 (pid, init_pid, path, create_if_missing, err);
+    {
+      ret = enter_cgroup_v2 (pid, init_pid, path, create_if_missing, err);
+      if (UNLIKELY (ret < 0))
+        return ret;
+    }
+  else
+    {
+      ret = enter_cgroup_v1 (pid, path, create_if_missing, err);
+      if (UNLIKELY (ret < 0))
+        return ret;
+    }
+  /* Reset the inherited cpu affinity. Old kernels do that automatically, but
+     new kernels remember the affinity that was set before the cgroup move.
+     This is undesirable, because it inherits the systemd affinity when the container
+     should really move to the container space cpus.
 
-  return enter_cgroup_v1 (pid, path, create_if_missing, err);
+     The sched_setaffinity call will always return an error (EINVAL or ENODEV)
+     when used like this. This is expected and part of the backward compatibility.
+
+     See: https://issues.redhat.com/browse/OCPBUGS-15102   */
+  ret = sched_setaffinity (pid, 0, NULL);
+  if (LIKELY (ret < 0))
+    {
+      if (UNLIKELY (! ((errno == EINVAL) || (errno == ENODEV))))
+        return crun_make_error (err, errno, "failed to reset affinity");
+    }
+  return 0;
 }
