@@ -31,6 +31,7 @@
 #include <signal.h>
 #include <sys/vfs.h>
 #include <inttypes.h>
+#include <fcntl.h>
 #include <time.h>
 
 #ifdef HAVE_SYSTEMD
@@ -68,17 +69,78 @@ get_systemd_scope_and_slice (const char *id, const char *cgroup_path, char **sco
     }
 }
 
+/* set the rt-runtime for the current cgroup and its parent if the path is not a scope.  */
+static int
+setup_rt_runtime (runtime_spec_schema_config_linux_resources *resources,
+                  const char *path, libcrun_error_t *err)
+{
+  cleanup_free char *cgroup_path = NULL;
+  cleanup_close int dirfd = -1;
+  bool need_set_parent = true;
+  char fmt_buf[64];
+  size_t len;
+  int ret;
+
+  if (resources == NULL || resources->cpu == NULL)
+    return 0;
+
+  if (has_suffix (path, ".scope"))
+    need_set_parent = false;
+
+  ret = append_paths (&cgroup_path, err, CGROUP_ROOT, "cpu", path, NULL);
+  if (UNLIKELY (ret < 0))
+    return ret;
+
+  dirfd = open (cgroup_path, O_DIRECTORY | O_CLOEXEC);
+  if (UNLIKELY (dirfd < 0))
+    return crun_make_error (err, errno, "open `%s`", cgroup_path);
+
+  if (resources->cpu->realtime_period)
+    {
+      len = sprintf (fmt_buf, "%" PRIu64, resources->cpu->realtime_period);
+
+      if (need_set_parent)
+        {
+          ret = write_file_at_with_flags (dirfd, O_WRONLY, 0, "../cpu.rt_period_us", fmt_buf, len, err);
+          if (UNLIKELY (ret < 0))
+            return ret;
+        }
+
+      ret = write_file_at_with_flags (dirfd, O_WRONLY, 0, "cpu.rt_period_us", fmt_buf, len, err);
+      if (UNLIKELY (ret < 0))
+        return ret;
+    }
+
+  if (resources->cpu->realtime_runtime)
+    {
+      len = sprintf (fmt_buf, "%" PRIu64, resources->cpu->realtime_runtime);
+
+      if (need_set_parent)
+        {
+          ret = write_file_at_with_flags (dirfd, O_WRONLY, 0, "../cpu.rt_runtime_us", fmt_buf, len, err);
+          if (UNLIKELY (ret < 0))
+            return ret;
+        }
+
+      ret = write_file_at_with_flags (dirfd, O_WRONLY, 0, "cpu.rt_runtime_us", fmt_buf, len, err);
+      if (UNLIKELY (ret < 0))
+        return ret;
+    }
+  return 0;
+}
+
 static int
 systemd_finalize (struct libcrun_cgroup_args *args, char **path_out,
                   int cgroup_mode, const char *suffix, libcrun_error_t *err)
 {
+  runtime_spec_schema_config_linux_resources *resources = args->resources;
+  cleanup_free char *cgroup_path = NULL;
   cleanup_free char *content = NULL;
   cleanup_free char *path = NULL;
   pid_t pid = args->pid;
   int ret;
   char *from, *to;
   char *saveptr = NULL;
-  cleanup_free char *cgroup_path = NULL;
 
   xasprintf (&cgroup_path, "/proc/%d/cgroup", pid);
   ret = read_all_file (cgroup_path, &content, NULL, err);
@@ -103,7 +165,7 @@ systemd_finalize (struct libcrun_cgroup_args *args, char **path_out,
 
       to = strchr (from, '\n');
       if (UNLIKELY (to == NULL))
-        return crun_make_error (err, 0, "cannot parse /proc/self/cgroup");
+        return crun_make_error (err, 0, "cannot parse `%s`", PROC_SELF_CGROUP);
       *to = '\0';
       if (suffix == NULL)
         path = xstrdup (from);
@@ -148,6 +210,11 @@ systemd_finalize (struct libcrun_cgroup_args *args, char **path_out,
                 }
             }
         }
+
+      ret = setup_rt_runtime (resources, path, err);
+      if (UNLIKELY (ret < 0))
+        return ret;
+
       break;
 
     case CGROUP_MODE_UNIFIED:
@@ -161,7 +228,7 @@ systemd_finalize (struct libcrun_cgroup_args *args, char **path_out,
         from += 3;
         to = strchr (from, '\n');
         if (UNLIKELY (to == NULL))
-          return crun_make_error (err, 0, "cannot parse /proc/self/cgroup");
+          return crun_make_error (err, 0, "cannot parse `%s`", PROC_SELF_CGROUP);
         *to = '\0';
         if (suffix == NULL)
           path = xstrdup (from);
@@ -202,7 +269,7 @@ systemd_finalize (struct libcrun_cgroup_args *args, char **path_out,
       break;
 
     default:
-      return crun_make_error (err, 0, "invalid cgroup mode %d", cgroup_mode);
+      return crun_make_error (err, 0, "invalid cgroup mode `%d`", cgroup_mode);
     }
 
   *path_out = path;
@@ -235,7 +302,10 @@ systemd_job_removed (sd_bus_message *m, void *userdata, sd_bus_error *error arg_
     {
       d->terminated = 1;
       if (strcmp (result, "done") != 0)
-        crun_make_error (&d->err, 0, "error %s systemd unit `%s`: got `%s`", d->op, unit, result);
+        {
+          crun_make_error (&d->err, 0, "error `%s` systemd unit `%s`: got `%s`", d->op, unit, result);
+          return -1;
+        }
     }
   return 0;
 }
@@ -539,6 +609,40 @@ open_sd_bus_connection (sd_bus **bus, libcrun_error_t *err)
   return 0;
 }
 
+static int
+get_value_from_unified_map (runtime_spec_schema_config_linux_resources *resources, const char *name,
+                            uint64_t *value, libcrun_error_t *err)
+{
+  size_t i;
+
+  if (resources == NULL || resources->unified == NULL)
+    return 0;
+
+  for (i = 0; i < resources->unified->len; i++)
+    if (strcmp (resources->unified->keys[i], name) == 0)
+      {
+        errno = 0;
+        *value = (uint64_t) strtoll (resources->unified->values[i], NULL, 10);
+        if (UNLIKELY (errno))
+          return crun_make_error (err, errno, "invalid value for `%s`: %s", name,
+                                  resources->unified->values[i]);
+        return 1;
+      }
+  return 0;
+}
+
+static inline int
+get_memory_limit (runtime_spec_schema_config_linux_resources *resources, uint64_t *limit, libcrun_error_t *err)
+{
+  if (resources->memory && resources->memory->limit_present)
+    {
+      *limit = resources->memory->limit;
+      return 1;
+    }
+
+  return get_value_from_unified_map (resources, "memory.max", limit, err);
+}
+
 static inline int
 get_weight (runtime_spec_schema_config_linux_resources *resources, uint64_t *weight, libcrun_error_t *err)
 {
@@ -551,22 +655,7 @@ get_weight (runtime_spec_schema_config_linux_resources *resources, uint64_t *wei
       return 1;
     }
 
-  if (resources->unified)
-    {
-      size_t i;
-
-      for (i = 0; i < resources->unified->len; i++)
-        if (strcmp (resources->unified->keys[i], "cpu.weight") == 0)
-          {
-            errno = 0;
-            *weight = (uint64_t) strtoll (resources->unified->values[i], NULL, 10);
-            if (UNLIKELY (errno))
-              return crun_make_error (err, errno, "invalid value for `cpu.weight`: %s",
-                                      resources->unified->values[i]);
-            return 1;
-          }
-    }
-  return 0;
+  return get_value_from_unified_map (resources, "cpu.weight", weight, err);
 }
 
 static int
@@ -575,17 +664,28 @@ append_resources (sd_bus_message *m,
                   int cgroup_mode,
                   libcrun_error_t *err)
 {
+  uint64_t memory_limit;
   int sd_err;
+  int ret;
 
   if (resources == NULL)
     return 0;
 
+  ret = get_memory_limit (resources, &memory_limit, err);
+  if (UNLIKELY (ret < 0))
+    return ret;
+  if (ret)
+    {
+      sd_err = sd_bus_message_append (m, "(sv)", "MemoryLimit", "t", memory_limit);
+      if (UNLIKELY (sd_err < 0))
+        return crun_make_error (err, -sd_err, "sd-bus message append MemoryLimit");
+    }
+
   switch (cgroup_mode)
     {
     case CGROUP_MODE_UNIFIED:
       {
         uint64_t weight;
-        int ret;
 
         ret = get_weight (resources, &weight, err);
         if (UNLIKELY (ret < 0))
@@ -611,7 +711,7 @@ append_resources (sd_bus_message *m,
       break;
 
     default:
-      return crun_make_error (err, 0, "invalid cgroup mode %d", cgroup_mode);
+      return crun_make_error (err, 0, "invalid cgroup mode `%d`", cgroup_mode);
     }
 
   return 0;
@@ -884,6 +984,9 @@ libcrun_destroy_systemd_cgroup_scope (struct libcrun_cgroup_status *cgroup_statu
 
   ret = systemd_check_job_status (bus, &job_data, object, "removing", err);
 
+  /* In case of a failed unit, call reset-failed so systemd can remove it. */
+  reset_failed_unit (bus, scope);
+
 exit:
   if (bus)
     sd_bus_unref (bus);
@@ -1042,7 +1145,7 @@ libcrun_update_resources_systemd (struct libcrun_cgroup_status *cgroup_status,
   sd_err = sd_bus_message_append (m, "sb", cgroup_status->scope, 1);
   if (UNLIKELY (sd_err < 0))
     {
-      ret = crun_make_error (err, -ret, "sd-bus message append");
+      ret = crun_make_error (err, -sd_err, "sd-bus message append");
       goto exit;
     }
 
@@ -1071,6 +1174,13 @@ libcrun_update_resources_systemd (struct libcrun_cgroup_status *cgroup_status,
       goto exit;
     }
 
+  if (cgroup_mode != CGROUP_MODE_UNIFIED)
+    {
+      ret = setup_rt_runtime (resources, cgroup_status->path, err);
+      if (UNLIKELY (ret < 0))
+        goto exit;
+    }
+
   ret = 0;
 
 exit:
