@@ -21,7 +21,6 @@ import sys
 import os
 import tempfile
 import subprocess
-import time
 
 base_conf = """
 {
@@ -144,13 +143,25 @@ def parse_proc_status(content):
         r[k] = v.strip()
     return r
 
-def add_all_namespaces(conf, cgroupns=False):
+def add_all_namespaces(conf, cgroupns=False, userns=False, netns=True, ipcns=True, utsns=True, pidns=True,time=False):
     has = {}
     for i in conf['linux']['namespaces']:
         has[i['type']] = i['type']
-    namespaces = ['pid', 'user', 'ipc', 'uts', 'network']
+    namespaces = []
+    if pidns:
+        namespaces = namespaces + ['pid']
+    if utsns:
+        namespaces = namespaces + ['uts']
     if cgroupns:
         namespaces = namespaces + ["cgroup"]
+    if ipcns:
+        namespaces = namespaces + ["ipc"]
+    if userns:
+        namespaces = namespaces + ["user"]
+    if netns:
+        namespaces = namespaces + ["network"]
+    if time:
+        namespaces = namespaces + ["time"]
     for i in namespaces:
         if i not in has:
             conf['linux']['namespaces'].append({"type" : i})
@@ -184,17 +195,32 @@ def run_all_tests(all_tests, allowed_tests):
 def get_tests_root():
     return '%s/.testsuite-run-%d' % (os.getcwd(), os.getpid())
 
+def get_tests_root_status():
+    return os.path.join(get_tests_root(), "root")
+
+def get_init_path():
+    return os.path.abspath(os.getenv("INIT") or "tests/init")
+
 def get_crun_path():
     cwd = os.getcwd()
     return os.getenv("OCI_RUNTIME") or os.path.join(cwd, "crun")
 
 def run_and_get_output(config, detach=False, preserve_fds=None, pid_file=None,
-                       command='run', env=None, use_popen=False, hide_stderr=False,
-                       all_dev_null=False, id_container=None, relative_config_path="config.json"):
-    temp_dir = tempfile.mkdtemp(dir=get_tests_root())
+                       command='run', env=None, use_popen=False, hide_stderr=False, cgroup_manager='cgroupfs',
+                       all_dev_null=False, id_container=None, relative_config_path="config.json",
+                       chown_rootfs_to=None):
+
+    # Some tests require that the container user, which might not be the
+    # same user as the person running the tests, is able to resolve the full path
+    # to its own tree
+    if chown_rootfs_to is not None:
+        temp_dir = tempfile.mkdtemp()
+    else:
+        temp_dir = tempfile.mkdtemp(dir=get_tests_root())
+
     rootfs = os.path.join(temp_dir, "rootfs")
     os.makedirs(rootfs)
-    for i in ["usr/bin", "etc", "var", "lib", "lib64", "usr/share/zoneinfo/Europe"]:
+    for i in ["usr/bin", "etc", "var", "lib", "lib64", "usr/share/zoneinfo/Europe", "proc", "sys", "dev"]:
         os.makedirs(os.path.join(rootfs, i))
     with open(os.path.join(rootfs, "var", "file"), "w+") as f:
         f.write("file")
@@ -211,7 +237,7 @@ def run_and_get_output(config, detach=False, preserve_fds=None, pid_file=None,
         conf = json.dumps(config)
         config_file.write(conf)
 
-    init = os.getenv("INIT") or "tests/init"
+    init = get_init_path()
     crun = get_crun_path()
 
     os.makedirs(os.path.join(rootfs, "sbin"))
@@ -222,12 +248,25 @@ def run_and_get_output(config, detach=False, preserve_fds=None, pid_file=None,
     os.symlink("../usr/share/zoneinfo/Europe/Rome", os.path.join(rootfs, "etc/localtime"))
     os.symlink("../foo/bar/not/here", os.path.join(rootfs, "etc/not-existing"))
 
+    # Populate /etc/passwd inside container rootfs with users root and test for various test-cases.
+    passwd = open(os.path.join(rootfs, "usr/share/passwd"), "w")
+    passwd.writelines(["root:x:0:0:root:/root:/bin/bash", "\ntest:x:1000:1000:test:/var/empty:/bin/bash"])
+    passwd.close()
+    os.symlink("../usr/share/passwd", os.path.join(rootfs, "etc/passwd"))
+
+    if chown_rootfs_to is not None:
+        os.chown(temp_dir, chown_rootfs_to, chown_rootfs_to)
+        for root, dirs, files in os.walk(temp_dir):
+            for f in dirs + files:
+                os.chown(os.path.join(root, f), chown_rootfs_to, chown_rootfs_to, follow_symlinks=False)
+
     detach_arg = ['--detach'] if detach else []
     preserve_fds_arg = ['--preserve-fds', str(preserve_fds)] if preserve_fds else []
     pid_file_arg = ['--pid-file', pid_file] if pid_file else []
     relative_config_path = ['--config', relative_config_path] if relative_config_path else []
 
-    args = [crun, command] + relative_config_path + preserve_fds_arg + detach_arg + pid_file_arg + [id_container]
+    root = get_tests_root_status()
+    args = [crun, "--cgroup-manager", cgroup_manager, "--root", root, command] + relative_config_path + preserve_fds_arg + detach_arg + pid_file_arg + [id_container]
 
     stderr = subprocess.STDOUT
     if hide_stderr:
@@ -250,11 +289,22 @@ def run_and_get_output(config, detach=False, preserve_fds=None, pid_file=None,
         return subprocess.check_output(args, cwd=temp_dir, stderr=stderr, env=env, close_fds=False).decode(), id_container
 
 def run_crun_command(args):
-    cwd = os.getcwd()
+    root = get_tests_root_status()
     crun = get_crun_path()
-    args = [crun] + args
+    args = [crun, "--root", root] + args
     return subprocess.check_output(args, close_fds=False).decode()
 
+# Similar as run_crun_command but does not performs decode of output and relays error message for further matching
+def run_crun_command_raw(args):
+    root = get_tests_root_status()
+    crun = get_crun_path()
+    args = [crun, "--root", root] + args
+    return subprocess.check_output(args, close_fds=False, stderr=subprocess.STDOUT)
+
+def running_on_systemd():
+    with open('/proc/1/comm') as f:
+        return "systemd" in f.readline()
+
 def tests_main(all_tests):
     os.environ["LANG"] = "C"
     tests_root = get_tests_root()
@@ -264,6 +314,13 @@ def tests_main(all_tests):
     finally:
         shutil.rmtree(tests_root)
 
+def is_rootless():
+    if os.getuid() != 0:
+        return True
+    with open("/proc/self/uid_map") as f:
+        if "4294967295" in f.readline():
+            return False
+    return True
 
 def get_crun_feature_string():
     for i in run_crun_command(['--version']).split('\n'):
